/*
 *	BIRD -- Endianity Conversion
 *
 *	(c) 1999 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#ifndef _BIRD_ENDIAN_H_
#define _BIRD_ENDIAN_H_

/* hton[sl] and ntoh[sl] are defined here */

#include <sys/types.h>
#include <netinet/in.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#endif
