# BIRD {{ this.version }} release ({{ kind }})

*This is an autogenerated issue from `tools/release.py`.
Do not edit by hand unless you know what you are doing.
Just tick off what is done.*

## End of development

*Things seem to be done. Checking, gathering, last fixes.*

- [ ] Issue last call in the team chat.
- [X] Next version milestone exists: [{{ next.milestone.name }}]({{ next.milestone.url }})
{%- if kind == "minor" %}
- [X] First patch milestone: [{{ patch.milestone.name }}]({{ patch.milestone.url }})
{%- endif %}
- [ ] [Current milestone ({{ this.milestone.name }})]({{ this.milestone.url }}) empty
- [ ] Walk once again over closed issues and check that they are indeed merged into [{{ this.branch.name }}]({{ this.branch.url }})
{%- if kind == "patch" %}
- [ ] Walk once again over closed issues and check they are either applied on ({{ main.branch.name }}]({{ main.branch.url }}) or a linked issue is in [{{ main.milestone.name }}]({{ main.milestone.url }})
{%- endif %}
- [ ] Pipeline is green (netlab, pkgs, install) *TODO: find out how to automate this*
- [ ] Documentation build works *TODO: find out how to automate this*

## Code freeze

*We are now confident that everything is working. This is the process
of doing all the non-coding things related to release.*

- [ ] Text for announcements ready in [bird-notes](https://gitlab.nic.cz/labs/bird-notes/-/tree/master/release_notes/{{ this.version }}.md) *put a symlink there if multiple versions are being released*
  - [ ] customer info
  - [ ] mailing-list
  - [ ] PR {% if kind == "patch" %}(not needed unless security){% endif %}
  - [ ] socials
- [ ] NEWS and version update: `tools/release-commit patch`
  - [~] Commit pushed, linked to this exact issue *%%AUTO:NEWS:TICK%%*
- [~] Official package archives on website: %%AUTO:NEWS:BLOCKED%%
  - [ ] In bird-web: `make version-update PIPELINE=%%AUTO:NEWS:PIPELINE%%`
  - [ ] In bird-web: Commit and run `make version-update` to check
  - [ ] In bird-web: `git push origin`
  - [ ] [Check SRC TGZ download at the staging website](https://bird-web-cz-nic-labs-bird-a63d8cfd75ea76921ab3b6970da0bbd08b1dc.pages.office.nic.cz/download/bird-{{ this.version }}.tar.gz)
  - [ ] [Check DOC TGZ download at the staging website](https://bird-web-cz-nic-labs-bird-a63d8cfd75ea76921ab3b6970da0bbd08b1dc.pages.office.nic.cz/download/bird-doc-{{ this.version }}.tar.gz)
  - [ ] In bird-web: `git push production`
  - [ ] [Check SRC TGZ download at the production website](https://bird.nic.cz/download/bird-{{ this.version }}.tar.gz)
  - [ ] [Check DOC TGZ download at the production website](https://bird.nic.cz/download/bird-doc-{{ this.version }}.tar.gz)
- [ ] Mail to downstream mailing-list sent
- [ ] Mail to customers sent
- [ ] Debian / Ubuntu packages done *TODO: find out how to automate this*
- [ ] CentOS packages done *TODO: find out how to automate this*

## External release

*Everything is now prepared to be shipped and we need to execute it.*

- [~] TAG in BIRD repo added %%AUTO:TAG:TICK%%
- [ ] Pipeline still green after TAG
- [~] Update website %%AUTO:TAG:BLOCKED%%
  - [ ] Add a new item in news
  - [~] In bird-web: `make version-update PIPELINE=%%AUTO:TAG:PIPELINE%%`
  - [ ] Commit and run `make version-update` again to check
  - [ ] In bird-web: `./update-doc.py`
  - [ ] In bird-web: `git add www/doc/*`
  - [ ] In bird-web: `git commit -a`
  - [ ] In bird-web: `git push origin`
  - [ ] [Check documentation at the staging website](https://bird-web-cz-nic-labs-bird-a63d8cfd75ea76921ab3b6970da0bbd08b1dc.pages.office.nic.cz/doc/bird-{{ this.version }}.html)
  - [ ] [Check news at the staging website](https://bird-web-cz-nic-labs-bird-a63d8cfd75ea76921ab3b6970da0bbd08b1dc.pages.office.nic.cz/)
  - [ ] [Check download page at the staging website](https://bird-web-cz-nic-labs-bird-a63d8cfd75ea76921ab3b6970da0bbd08b1dc.pages.office.nic.cz/get-bird)
  - [ ] In bird-web: `git push production`
  - [ ] [Check documentation at the production website](https://bird.nic.cz/doc/bird-{{ this.version }}.html)
  - [ ] [Check news at the production website](https://bird.nic.cz/)
  - [ ] [Check download page at the production website](https://bird.nic.cz/get-bird)
- [ ] Announcement to bird-users mailing-list made
- [ ] PR department informed
- [ ] Boasted on socials (Mastodon, Linkedin)
- [ ] All relevant support tickets resolved / closed
- [ ] Post-release debrief done

