#
#	BIRD -- Internal Commands Of The Client
#
#	(c) 2000 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
#
#	Can be freely distributed and used under the terms of the GNU GPL.
#

CF_CLI(QUIT,,, [[Quit the client]])
CF_CLI(EXIT,,, [[Exit the client]])
CF_CLI(HELP,,, [[Description of the help system]])
